@@ -1,8 +1,17 @@
+# TODO Add warning for truncated dates!
+# TODO test standalone script
+# TODO add random access for gzip
+# TODO interleave skip files until their first epoch is reached
+# TODO Maybe Time::Piece would be a better fit than Date::Manip?
+# TODO output version
+
 use strict;
 use warnings;
 
 package App::dategrep;
 use Date::Manip::Date;
+use Date::Manip::Delta;
+use Config::Tiny;
 use Pod::Usage;
 use Getopt::Long;
 use Fcntl ":seek";
@@ -10,7 +19,7 @@ use File::Basename qw(basename);
 use base 'Exporter';
 our @EXPORT_OK = qw(run);
 
-our $VERSION = '0.15';
+our $VERSION = '0.50';
 
 our $app;
 
@@ -26,8 +35,21 @@ sub error {
     return $rc;
 }
 
-sub run {
+sub intervall_to_epoch {
+    my ( $time, $format ) = @_;
+    if ( $time =~ /^(.*) from (.*)$/ ) {
+        my ( $delta, $date ) =
+          ( Date::Manip::Delta->new($1), Date::Manip::Date->new($2) );
+        ## TODO: $date->is_date is missing in Date::Manip::Date
+        ## will be fixed in next major release
+        if ( $delta->is_delta() ) { ## and $date->is_date() ) {
+            return $date->calc($delta)->secs_since_1970_GMT();
+        }
+    }
+    return date_to_epoch( $time, $format );
+}
 
+sub run {
     my %options;
     if ( $ENV{DATEGREP_DEFAULT_FORMAT} ) {
         $options{format} = $ENV{DATEGREP_DEFAULT_FORMAT};
@@ -37,6 +59,7 @@ sub run {
         \%options,        'start|from=s', 'end|to=s',     'format=s',
         'last-minutes=i', 'multiline!',   'blocksize=i',  'help|?',
         'sort-files',     'man',          'configfile=s', 'interleave',
+        'byte-offsets',   'debug=s',
     );
     if ( !$rc ) {
         pod2usage( -exitstatus => "NOEXIT", -verbose => 0 );
@@ -55,6 +78,7 @@ sub run {
     my $config = loadconfig( $options{configfile} );
 
     my %named_formats = (
+        'iso8601' => "%O%Z",
         'rsyslog' => "%b %e %H:%M:%S",
         'apache'  => "%d/%b/%Y:%T %z",
     );
@@ -71,21 +95,15 @@ sub run {
         $options{'format'} = $named_formats{ $options{'format'} };
     }
 
-    my ( $start, $end, $error ) = ( 0, time() );
+    my ( $start, $end ) = ( 0, time() );
 
     if ( defined $options{'start'} ) {
-        ( $start ) = date_to_epoch( $options{'start'} );
-	if ( not defined $start ) {
-		( $start ) = date_to_epoch( $options{'start'}, $options{'format'} );
-	}
+        ($start) = intervall_to_epoch( $options{'start'}, $options{'format'} );
         return error("Illegal start time.") if not defined $start;
     }
 
     if ( defined $options{'end'} ) {
-        ( $end ) = date_to_epoch( $options{'end'} );
-	if ( not defined $end ) {
-		( $end ) = date_to_epoch( $options{'end'}, $options{'format'} );
-	}
+        ($end) = intervall_to_epoch( $options{'end'}, $options{'format'} );
         return error("Illegal end time.") if not defined $end;
     }
 
@@ -103,24 +121,43 @@ sub run {
         ( $start, $end ) = ( $end, $start );
     }
 
+    if ( defined $options{'debug'} && $options{'debug'} eq 'time' ) {
+        print "Start: $start End: $end\n";
+        return 0;
+    }
+
     if ( !@ARGV ) {
         push @ARGV, '-';
     }
 
     eval {
 
-        my @iters = map { get_iterator( $_, $start, $end, %options ) } @ARGV;
+        if ( $options{'byte-offsets'} ) {
+            if ( @ARGV == 1 and -f $ARGV[0] ) {
+                my ( $fh, $byte_beg, $byte_end ) =
+                  normal_file_byte_offsets( $ARGV[0], $start, $end, %options );
+                if ( not defined $byte_end ) {
+                    $byte_end = ( stat($fh) )[7];
+                }
+                print "$byte_beg $byte_end\n";
+                return 0;
+            }
+        }
+
+        my @iterators =
+          map { get_iterator( $_, $start, $end, %options ) } @ARGV;
+
 
         if ( $options{'interleave'} ) {
-            interleave_iterators( $options{'format'}, @iters );
+            interleave_iterators( $options{'format'}, @iterators );
             return 0;
         }
 
         if ( $options{'sort-files'} ) {
-            @iters = sort_iterators( $options{'format'}, @iters );
+            @iterators = sort_iterators( $options{'format'}, @iterators );
         }
 
-        for my $iter (@iters) {
+        for my $iter (@iterators) {
             if ($iter) {
                 while ( my $line = $iter->() ) {
                     print $line;
@@ -132,27 +169,45 @@ sub run {
     return 0;
 }
 
-sub interleave_iterators {
-    my ( $format, @iters ) = @_;
-    ## TODO choices is a bad name, better ideas?  ## TODO my
-    ## TODO duplicate code
-    my @choices;
-    for my $iter (@iters) {
-        my $line = $iter->();
-        next if !$line;
-        my $epoch = date_to_epoch( $line, $format );
-        push @choices, [ $epoch, $line, $iter ];
-    }
-    @choices = sort { $a->[0] <=> $b->[0] } @choices;
-    while ( my $choice = shift @choices ) {
-        my ( $epoch, $line, $iter ) = @$choice;
-        print $line;
+=pod 
 
-        $line = $iter->();
-        next if !$line;
-        $epoch = date_to_epoch( $line, $format );
-        @choices =
-          sort { $a->[0] <=> $b->[0] } @choices, [ $epoch, $line, $iter ];
+=item guess_format( $formats, @iterators )
+
+Check all formats in the array reference $formats against the first
+line of all iterators. Return the first that matched.
+
+=cut
+
+sub guess_format {
+    my ($formats, @iterators) = @_;
+    for my $iterator (@iterators) {
+        my $line = $iterator->( peek => 1 );
+        for my $format ( @$formats ) {
+            my $epoch = date_to_epoch( $line, $format );
+            if ( defined $epoch ) {
+                return $format;
+            }
+        }
+    }
+    return;
+}
+
+=pod
+
+=item interleave_iterators( $format, @iterators )
+
+Take a list of iterators and checks every iterator for its next
+line. After sorting these lines according to their dates, print the
+earliest line. I<$format> is the date specification to find dates in
+lines and @iterators a list of iterators produced by I<get_iterator()>.
+
+=cut
+
+sub interleave_iterators {
+    my ( $format, @iterators ) = @_;
+
+    while ( @iterators = sort_iterators( $format, @iterators ) ) {
+        print $iterators[0]->();
     }
     return;
 }
@@ -161,7 +216,7 @@ sub get_iterator {
     my ( $filename, $start, $end, %options ) = @_;
     my $iter;
     if ( $filename eq '-' ) {
-        $iter = stdin_iterator( $filename, $start, $end, %options );
+        $iter = stdin_iterator( $start, $end, %options );
     }
     elsif ( $filename =~ /\.(bz|bz2|gz|z)$/ ) {
         $iter = uncompress_iterator( $filename, $start, $end, %options );
@@ -184,13 +239,30 @@ sub get_iterator {
     };
 }
 
+=pod
+
+=item sort_iterators( $format, @iterators )
+
+Take a date format and a list of iterators and return a list of
+iterators sorted by the date of their first lines. If an iterators
+returns no line, it is not included in the returned list.
+
+=cut
+
 sub sort_iterators {
     my ( $format, @iterators ) = @_;
+
     my @timestamps;
     for my $iterator (@iterators) {
         my $line = $iterator->( peek => 1 );
+        
+        ## remove all iterators with eof
+        next if not defined $line;
+
+        ## TODO What should we do under --multiline?
         my ( $epoch, $error ) = date_to_epoch( $line, $format );
         if ( !$epoch ) {
+            ## TODO Which iterator produced the error?
             die "No date found in first line: $error\n";
         }
         push @timestamps, [ $epoch, $iterator ];
@@ -198,83 +270,122 @@ sub sort_iterators {
     return map { $_->[1] } sort { $a->[0] <=> $b->[0] } @timestamps;
 }
 
-sub normal_file_iterator {
+sub normal_file_byte_offsets {
     my ( $filename, $start, $end, %options ) = @_;
 
     open( my $fh, '<', $filename ) or die "Can't open $filename: $!\n";
-    my $test_line = <$fh>;
-    if ( defined($test_line) ) {
-        my ( $epoch, $error ) =
-          date_to_epoch( $test_line, $options{'format'} );
-        if ($error) {
-            die "No date found in first line: $error\n";
-        }
-        seek( $fh, 0, SEEK_SET );
+    my $tell_beg = search(
+        $fh, $start, $options{'format'},
+        multiline => $options{multiline},
+        blocksize => $options{blocksize},
+    );
 
-        my $tell_beg = search(
-            $fh, $start, $options{'format'},
+    if ( defined $tell_beg ) {
+        my $tell_end = search(
+            $fh, $end, $options{'format'},
+            min_byte  => $tell_beg,
             multiline => $options{multiline},
             blocksize => $options{blocksize},
         );
 
-        if ( defined $tell_beg ) {
-            my $tell_end = search(
-                $fh, $end, $options{'format'},
-                min_byte  => $tell_beg,
-                multiline => $options{multiline},
-                blocksize => $options{blocksize},
-            );
-
-            seek( $fh, $tell_beg, SEEK_SET );
-
-            return sub {
-                my $line = <$fh>;
-                return if defined($tell_end) && ( tell() > $tell_end );
-                return $line;
-            };
-        }
+        return $fh, $tell_beg, $tell_end;
     }
     return;
 }
 
+
+=pod 
+
+=item normal_file_iterator( $filename, $start, $end, %options )
+
+Return a iterator subroutine, that will return lines from I<$filename>
+starting at I<$start> to I<$end> every time it is called. I<$start>
+end <$end> are timevalues in epoch.
+
+=cut
+
+sub normal_file_iterator {
+    my ( $filename, $start, $end, %options ) = @_;
+    my ( $fh, $tell_beg, $tell_end ) = normal_file_byte_offsets(@_);
+    if ( defined($tell_beg) ) {
+        seek( $fh, $tell_beg, SEEK_SET );
+        return sub {
+            my $line = <$fh>;
+            ## TODO can $tell_end be undefined?
+            return if defined($tell_end) && ( tell() > $tell_end );
+            return $line;
+        };
+    }
+    return;
+}
+
+=pod 
+
+=item uncompress_iterator( $filename, $start, $end, %options )
+
+Return a iterator subroutine, that will return lines from I<$filename>
+starting at I<$start> to I<$end> every time it is called. I<$start>
+end <$end> are timevalues in epoch. I<$filename> must be a gzip or
+bzip2 compressed file with a corresponding suffix of gz and z or bz
+and bzip2.
+
+=cut
+
 sub uncompress_iterator {
     my ( $filename, $start, $end, %options ) = @_;
-    my $uncompress;
+    my @uncompress;
     if ( $filename =~ /\.(bz|bz2)$/ ) {
-        $uncompress = 'bzcat';
+        @uncompress = qw(bzcat);
     }
     elsif ( $filename =~ /\.(gz|z)$/ ) {
-        $uncompress = 'zcat';
+        @uncompress = qw(gzip -c -d);
     }
     else {
-        die "unknown ending for compressed file\n";
+        die "unknown ending for compressed file $filename\n";
     }
-    open( my $pipe, '-|', $uncompress, $filename )
-      or die "Can't open $uncompress: $!\n";
+    open( my $pipe, '-|', @uncompress, $filename )
+      or die "Can't open @uncompress: $!\n";
     return fh_iterator( $pipe, $start, $end, %options );
 }
 
 sub stdin_iterator {
-    my ( $filename, $start, $end, %options ) = @_;
+    my ( $start, $end, %options ) = @_;
     return fh_iterator( \*STDIN, $start, $end, %options );
 }
 
 sub fh_iterator {
     my ( $fh, $start, $end, %options ) = @_;
     my $last_epoch = 0;
+
+    ## when we find the first line that was logged at $end, we
+    ## just return undef and set $found_end to one. We check
+    ## $found_end directly at the beginning of the iterator
+    ## function. If its true, we just return undef without
+    ## checking the date of the line.
+
+    my $found_end;
+
     return sub {
-      LINE: while ( my $line = <$fh> ) {
-            my ( $epoch, $error ) =
-              date_to_epoch( $line, $options{'format'} );
+      LINE:
+        while ( my $line = <$fh> ) {
+            return if $found_end;
+            my ( $epoch, $error ) = date_to_epoch( $line, $options{'format'} );
             if ( !$epoch ) {
-                if ( $options{'multiline'} ) {
-                    return $line if $last_epoch >= $start;
+                if ( $options{'multiline'} and $last_epoch >= $start ) {
+                    return $line;
                 }
                 die "Unparsable line: $line\n";
             }
             next LINE if $epoch < $start;
+
             $last_epoch = $epoch;
-            return if $epoch >= $end;
+
+            if ( $epoch >= $end ) {
+                ## see comment above
+                $found_end = 1;
+                return;
+            }
+
             if ( $epoch >= $start ) {
                 return $line;
             }
@@ -292,26 +403,11 @@ sub loadconfig {
         return;
     }
 
-    my ( %config, $section );
-    open( my $cfg_fh, '<', $configfile )
-      or die "Can't open config file: $!\n";
-    while (<$cfg_fh>) {
-        next if /^\s*\#/ || /^\s*$/;
-        if (/^\[([^\]]*)\]\s*$/) {
-            $section = lc $1;
-        }
-        elsif (/^(\w+)\s*=\s*(.*)/) {
-            my ( $key, $val ) = ( $1, $2 );
-            if ( not defined $section ) {
-                die "parameter $key not in section\n";
-            }
-            $config{$section}->{$key} = $val;
-        }
-        else {
-            die "Parse error in configuration file\n";
-        }
+    my $config = Config::Tiny->read( $configfile );
+    if ( not defined $config ) {
+        die "Error while parsing configfile: " . Config::Tiny->errstr() . "\n";
     }
-    return \%config;
+    return $config;
 }
 
 {
@@ -322,13 +418,17 @@ sub loadconfig {
         if ( !$date ) {
             $date = Date::Manip::Date->new();
         }
-        my $error =
-          defined($format)
-          ? $date->parse_format( $format, $str )
-          : $date->parse($str);
-        if ($error) {
-            return ( undef, $date->err );
+
+        my $error;
+        if ($format) {
+            $error = $date->parse_format( $format, $str );
         }
+
+        if ( !$format or $error ) {
+            $error = $date->parse($str);
+        }
+
+        return ( undef, $date->err ) if $error;
         return ( $date->secs_since_1970_GMT() );
     }
 }
